import os
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

from tktl.core.exceptions import TaktileSdkError
from tktl.core.loggers import LOG
from tktl.core.schemas.repository import Endpoint, Repository, RepositoryDeployment
from tktl.core.t import Resources


def default_comp(x: Any, y: Any):
    return x == y


def branch_comp(x: str, y: str):
    return x == y or os.path.basename(y) == x


def is_in_sequence_comp(x: Sequence, y: str):
    return y in set(x)


def filter_prop(
    resources: Sequence[Resources],
    prop_name: str,
    value: Any,
    key: Callable = default_comp,
):
    if value is not None:
        return [d for d in resources if key(value, getattr(d, prop_name))]
    return resources


def filter_deployments(
    deployments: List[RepositoryDeployment],
    git_hash: str = None,
    branch_name: str = None,
    status_name: Union[str, List[str]] = None,
) -> List[RepositoryDeployment]:
    for n, v in zip(
        ["commit_hash", "git_ref", "status"], [git_hash, branch_name, status_name]
    ):
        if n == "git_ref":
            comp = branch_comp
        elif n == "commit_hash" and git_hash:
            if len(git_hash) < 40:

                def comp(x: str, y: str):
                    return y.startswith(x)

            else:
                comp = default_comp
        elif n == "status":
            if isinstance(status_name, list):
                comp = is_in_sequence_comp
            else:
                comp = default_comp
        else:
            comp = default_comp
        deployments = filter_prop(deployments, prop_name=n, value=v, key=comp)

    return deployments


def filter_repositories(
    repositories: List[Repository],
    repository_name: str = None,
    repository_owner: str = None,
) -> List[Repository]:
    for n, v in zip(
        ["repository_name", "repository_owner"], [repository_name, repository_owner]
    ):
        repositories = filter_prop(repositories, prop_name=n, value=v)
    return repositories


def filter_endpoints(
    endpoints: List[Endpoint], endpoint_name: str = None, endpoint_kind: str = None
) -> List[Endpoint]:

    for n, v in zip(["name", "kind"], [endpoint_name, endpoint_kind]):
        endpoints = filter_prop(endpoints, prop_name=n, value=v)
    return endpoints


def get_deployment_from_endpoint_and_branch(
    endpoint_mapping: List[Dict[str, Union[Endpoint, RepositoryDeployment]]],
    endpoint_name: str,
    branch_name: str = None,
) -> Optional[RepositoryDeployment]:
    filtered = [
        e
        for e in endpoint_mapping
        if e and cast(Endpoint, e["endpoint"]).name == endpoint_name
    ]
    if len(filtered) == 0:
        LOG.warning(
            f"No endpoints with name: {endpoint_name} found across all deployed branches"
        )
        return None
    if len(filtered) == 1:
        deployment = cast(RepositoryDeployment, filtered[0]["deployment"])
        if branch_name is not None and not branch_comp(deployment.git_ref, branch_name):
            LOG.warning(
                f"No endpoint {endpoint_name} for branch {branch_name}, but found one for branch: "
                f"{os.path.basename(deployment.git_ref)}. Returning that instead"
            )
        return deployment
    else:
        available_branches = set(
            [
                os.path.basename(cast(RepositoryDeployment, e["deployment"]).git_ref)
                for e in endpoint_mapping
            ]
        )
        if branch_name is None:
            LOG.warning(
                f"Ambiguous: more than one branch for repository with endpoint: {endpoint_name}. Please "
                f"specify a branch. Available branches: with such endpoint: {', '.join(available_branches)}"
            )
            return None
        branch_filtered = [
            e
            for e in filtered
            if branch_comp(
                cast(RepositoryDeployment, e["deployment"]).git_ref, branch_name
            )
        ]
        if len(branch_filtered) == 0:
            LOG.warning(
                f"No endpoints with name: {endpoint_name} and branch {branch_name} found. Available "
                f"branches with such endpoint: {', '.join(available_branches)}"
            )
            return None
        elif len(branch_filtered) == 1:
            return cast(RepositoryDeployment, branch_filtered[0]["deployment"])
        else:
            print(
                [
                    (
                        cast(RepositoryDeployment, e["deployment"]).git_ref,
                        cast(RepositoryDeployment, e["deployment"]).status,
                    )
                    for e in endpoint_mapping
                ]
            )
            # Impossible to have more than one endpoint in the same branch for the same repo with the same name
            raise TaktileSdkError("This should never, ever happen")



from docker.errors import APIError  # type: ignore
from pydantic import ValidationError

from tktl.core.exceptions.exceptions import (
    MissingDocker,
    NoContentsFoundException,
    UserRepoValidationException,
)
from tktl.core.loggers import LOG
from tktl.core.managers.docker import DockerManager
from tktl.core.managers.project import ProjectManager
from tktl.core.schemas.project import ProjectValidationOutput
from tktl.core.validation.outputs import (
    ConfigFileValidationFailure,
    ProjectValidationFailure,
)


def validate_project_config(path: str):
    try:
        ProjectManager.validate_project_config(path)
    except ValidationError as e:
        validation_output = ProjectValidationOutput(
            title=ConfigFileValidationFailure.title,
            summary=ConfigFileValidationFailure.summary,
            text=ConfigFileValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    except (NoContentsFoundException, UserRepoValidationException) as e:
        validation_output = ProjectValidationOutput(
            title=ProjectValidationFailure.title,
            summary=ProjectValidationFailure.summary,
            text=ProjectValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    LOG.log("Project scaffolding is valid!", color="green")


def log_failure(validation_output: ProjectValidationOutput):
    LOG.log(f"Project scaffolding is invalid: {validation_output.title}", color="red")
    LOG.log(validation_output.summary, color="red", err=True)
    LOG.log(validation_output.text, color="red", err=True)


def build_image(path: str, nocache: bool = False):
    dm = DockerManager(path)

    LOG.log("Patching docker file...")
    dm.patch_docker_file()
    LOG.log("Building docker image...")
    image = dm.build_image(nocache=nocache)
    dm.remove_patched_docker_file()

    return dm, image


def status_output(status, logs):
    if status["StatusCode"] == 0:
        LOG.log("Success", color="green")
    else:
        LOG.log("Error", color="red")
        print(logs.decode())


def validate_import(path: str, nocache: bool):
    try:
        dm, image = build_image(path, nocache=nocache)
        LOG.log("Testing import of tktl...")
        status, logs = dm.test_import(image)
        status_output(status, logs)
    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_unittest(path: str, nocache: bool):
    try:
        dm, image = build_image(path=path, nocache=nocache)
        LOG.log("Running unittests...")
        status, logs = dm.test_unittest(image)
        status_output(status, logs)
    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_integration(path: str, nocache: bool, timeout: int, retries: int):
    try:
        dm, image = build_image(path, nocache=nocache)
        LOG.log("Waiting for service to start...")
        rest_response, grpc_response, container = dm.run_and_check_health(
            image,
            kill_on_success=True,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
        )
        if _validate_container_response(
            rest_response=rest_response,
            grpc_response=grpc_response,
            output=container.logs(),
        ):
            LOG.log("Success", color="green")
        else:
            LOG.log(
                "Unable to run container. See stack trace for more info", color="red"
            )

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_profiling(path: str, nocache: bool, timeout: int, retries: int):
    try:
        dm, image = build_image(path, nocache=nocache)
        LOG.log("Testing profiling...")
        LOG.log("Initiating service...")
        rest_response, grpc_response, container = dm.run_and_check_health(
            image,
            kill_on_success=False,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
        )
        success = _validate_container_response(
            rest_response=rest_response,
            grpc_response=grpc_response,
            output=container.logs(),
        )
        if not success:
            LOG.log(
                "Failed to run service container, ensure service can run with `tktl validate integration`",
                color="red",
            )
            return
        try:
            LOG.log("Initiating remote profiling...")
            status, logs = dm.run_profiling_container()
            status_output(status, logs)
        finally:
            try:
                container.kill()
            except APIError:
                pass

    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def _validate_container_response(rest_response, grpc_response, output):
    if rest_response is None:
        LOG.log("Could not access REST endpoint", color="red")
        print(output.decode())
    elif rest_response.status_code != 200:
        LOG.log(f"Response status code {rest_response.status_code}", color="red")
        print(output.decode())
    elif grpc_response is None:
        LOG.log("Could not access gRPC endpoint", color="red")
    else:
        return True
    return False


def validate_all(path: str, nocache: bool, timeout: int, retries: int):
    validate_project_config(path)
    validate_import(path, nocache=nocache)
    validate_unittest(path, nocache=False)
    validate_integration(path, nocache=False, timeout=timeout, retries=retries)
    validate_profiling(path, nocache=False, timeout=timeout, retries=retries)
